const axios = require('axios');
const ethers = require('ethers');
const crypto = require('crypto');
const objectHash = require('object-hash');
const Promise = require('bluebird');

require('dotenv').config();

const verifySignature = require('./verify-signature');
const { getContractAddressByNetworkName } = require('./contract-config');
const { authorAddresses, authorPrivKeys, forum } = require('./db');
const DAOArtifact = require('../contractArtifacts/DAO.json');

const network = process.env.ETH_NETWORK;
console.log('network:', network);
const getProvider = () => {
  switch (network) {
    case 'localhost':
      return ethers.getDefaultProvider('http://localhost:8545');
    case 'sepolia':
      return new ethers.InfuraProvider(
        network,
        process.env.INFURA_API_KEY,
      );
    default:
      throw new Error('Unknown network');
  }
};
const signer = new ethers.Wallet(process.env.ETH_PRIVATE_KEY, getProvider());
const getContract = (name) => new ethers.Contract(
  getContractAddressByNetworkName(process.env.ETH_NETWORK, name),
  DAOArtifact.abi,
  signer,
);

const fetchPaperInfo = async (paperId, retryDelay = 5000) => {
  const url = `https://api.semanticscholar.org/graph/v1/paper/${paperId}?fields=title,url,authors,references`;
  let retry = false;
  let paper;
  const response = await axios.get(url, {
    headers: {
      'api-key': process.env.SEMANTIC_SCHOLAR_API_KEY,
    },
  }).catch(async (error) => {
    if (error.response?.status === 429) {
      // Rate limit
      retry = true;
      return;
    }
    // Some other error occurred
    throw new Error(error);
  });
  if (retry) {
    console.log('retry delay (sec):', retryDelay / 1000);
    await new Promise((resolve) => {
      setTimeout(resolve, retryDelay);
    });
    paper = await fetchPaperInfo(paperId, retryDelay * 2);
  } else {
    paper = response.data;
  }
  return paper;
};

const getAuthorsInfo = async (paper) => Promise.mapSeries(
  paper.authors.filter((x) => !!x.authorId),
  async ({ authorId }) => {
  // Check if we already have an account for each author
    let authorAddress;
    let authorPrivKey;
    try {
      authorAddress = await authorAddresses.get(authorId);
    } catch (e) {
    // Probably not found
    }
    if (authorAddress) {
    // This should always succeed, so we don't use try/catch here
      authorPrivKey = await authorPrivKeys.get(authorAddress);
    } else {
    // Generate and store a new account
      const id = crypto.randomBytes(32).toString('hex');
      authorPrivKey = `0x${id}`;
      const wallet = new ethers.Wallet(authorPrivKey);
      authorAddress = wallet.address;
      await authorAddresses.put(authorId, authorAddress);
      await authorPrivKeys.put(authorAddress, authorPrivKey);
    }
    return {
      authorAddress,
      authorPrivKey,
    };
  },
);

const generatePost = async (paper) => {
  const authorsInfo = await getAuthorsInfo(paper);
  if (!authorsInfo.length) {
    throw new Error('Paper has no authors with id');
  }
  const firstAuthorWallet = new ethers.Wallet(authorsInfo[0].authorPrivKey);
  const eachAuthorWeightPercent = Math.floor(100 / authorsInfo.length);
  const authors = authorsInfo.map(({ authorAddress }) => ({
    weightPercent: eachAuthorWeightPercent,
    authorAddress,
  }));
    // Make sure author weights sum to 100
  const totalAuthorsWeight = authors.reduce((t, { weightPercent }) => t + weightPercent, 0);
  authors[0].weightPercent += 100 - totalAuthorsWeight;

  const content = `Semantic Scholar paper ${paper.paperId}
${paper.title}
HREF ${paper.url}`;

  const embeddedData = {
    semanticScholarPaperId: paper.paperId,
  };
  let contentToSign = content;
  if (embeddedData && Object.entries(embeddedData).length) {
    contentToSign += `\n\n${JSON.stringify(embeddedData, null, 2)}`;
  }
  const signature = firstAuthorWallet.signMessageSync(contentToSign);
  const verified = verifySignature({
    authors, content, signature, embeddedData,
  });
  if (!verified) {
    throw new Error('Signature verification failed');
  }
  const hash = objectHash({
    authors, content, signature, embeddedData,
  });
  return {
    hash, authors, content, signature, embeddedData,
  };
};

module.exports = async (req, res) => {
  const dao = await getContract('DAO');
  const {
    body: {
      paperId,
    },
  } = req;

  console.log(`importFromSS ${paperId}`);

  // Read the paper info from SS
  const paper = await fetchPaperInfo(paperId);

  // Send HTTP code 202 - Accepted
  res.status(202).end();

  console.log('references count:', paper.references.length);

  const eachCitationWeightPercent = Math.floor(30 / paper.references.length);
  const citations = await Promise.mapSeries(
    paper.references.filter((x) => !!x.paperId),
    async ({ paperId: citedPaperId }) => {
      // We need to fetch this paper so we can generate the post we WOULD add to the forum.
      // That way, if we later add the cited paper to the blockchain it will have the correct hash.
      // The forum allows dangling citations to support this use case.
      const citedPaper = await fetchPaperInfo(citedPaperId);
      const citedPost = await generatePost(citedPaper);
      return {
        weightPercent: eachCitationWeightPercent,
        targetPostId: citedPost.hash,
      };
    },
  );

  // Make sure citation weights sum to 100
  const totalCitationWeight = citations.reduce((t, { weightPercent }) => t + weightPercent, 0);
  citations[0].weightPercent += 100 - totalCitationWeight;

  // Create a post for this paper
  const {
    hash, authors, content, signature, embeddedData,
  } = await generatePost(paper);

  // Write the new post to our database
  await forum.put(hash, {
    authors, content, signature, embeddedData, citations,
  });

  // Add the post to the form (on-chain)
  try {
    await dao.addPost(authors, hash, citations);
  } catch (e) {
    if (e.reason === 'A post with this contentId already exists') {
      console.log(`Post already added for paper ${paperId}`);
      return;
    }
    throw e;
  }

  console.log(`Added post to blockchain for paper ${paperId}`);
};
