// SPDX-License-Identifier: Unlicense
pragma solidity ^0.8.24;

import "./core/DAO.sol";
import "./core/Forum.sol";
import "./Proposals.sol";
import "./interfaces/IAcceptAvailability.sol";
import "./interfaces/IOnProposalAccepted.sol";
import "hardhat/console.sol";

abstract contract WorkContract is
    DAOContract,
    IAcceptAvailability,
    IOnProposalAccepted
{
    struct AvailabilityStake {
        address worker;
        uint256 amount;
        uint endTime;
        bool assigned;
    }

    enum WorkStatus {
        Requested,
        EvidenceSubmitted,
        ApprovalSubmitted,
        Complete
    }

    struct WorkRequest {
        address customer;
        uint256 fee;
        WorkStatus status;
        uint stakeIndex;
        string requestContentId;
        string evidenceContentId;
        Citation[] citations;
        bool approval;
    }

    struct PriceProposal {
        uint price;
        uint proposalIndex;
    }

    Proposals proposalsContract;
    uint public price;
    mapping(uint => PriceProposal) public priceProposals;
    uint public priceProposalCount;
    mapping(uint => AvailabilityStake) public stakes;
    uint public stakeCount;
    mapping(uint => WorkRequest) public requests;
    uint public requestCount;

    uint constant POOL_DURATION = 20;

    event AvailabilityStaked(uint stakeIndex);
    event WorkAssigned(uint requestIndex, uint stakeIndex);
    event WorkEvidenceSubmitted(uint requestIndex);
    event WorkApprovalSubmitted(uint requestIndex, bool approval);
    event PriceChangeProposed(uint priceProposalIndex);
    event PriceChangeAccepted(uint priceProposalIndex, uint price);

    constructor(
        DAO dao,
        Proposals proposalsContract_,
        uint price_
    ) DAOContract(dao) {
        price = price_;
        proposalsContract = proposalsContract_;
    }

    /// Accept availability stakes as reputation token transfer
    function acceptAvailability(
        address sender,
        uint256 amount,
        uint duration
    ) external {
        require(amount > 0, "No stake provided");
        uint stakeIndex = stakeCount++;
        AvailabilityStake storage stake = stakes[stakeIndex];
        stake.worker = sender;
        stake.amount = amount;
        stake.endTime = block.timestamp + duration;
        emit AvailabilityStaked(stakeIndex);
    }

    function extendAvailability(uint stakeIndex, uint duration) external {
        AvailabilityStake storage stake = stakes[stakeIndex];
        require(
            msg.sender == stake.worker,
            "Worker can only extend their own availability stake"
        );
        require(!stake.assigned, "Stake has already been assigned work");
        if (block.timestamp > stake.endTime) {
            stake.endTime = block.timestamp + duration;
        } else {
            stake.endTime = stake.endTime + duration;
        }
        emit AvailabilityStaked(stakeIndex);
    }

    /// Select a worker randomly from among the available workers, weighted by amount staked
    function randomWeightedSelection() internal view returns (uint stakeIndex) {
        uint totalStakes;
        for (uint i = 0; i < stakeCount; i++) {
            if (stakes[i].assigned) continue;
            if (block.timestamp > stakes[i].endTime) continue;
            totalStakes += stakes[i].amount;
        }
        require(totalStakes > 0, "No available worker stakes");
        uint select = block.prevrandao % totalStakes;
        uint acc;
        for (uint i = 0; i < stakeCount; i++) {
            if (stakes[i].assigned) continue;
            if (block.timestamp > stakes[i].endTime) continue;
            acc += stakes[i].amount;
            if (acc > select) {
                stakeIndex = i;
                break;
            }
        }
    }

    /// Assign a random available worker
    function assignWork(uint requestIndex) internal returns (uint stakeIndex) {
        stakeIndex = randomWeightedSelection();
        AvailabilityStake storage stake = stakes[stakeIndex];
        stake.assigned = true;
        emit WorkAssigned(requestIndex, stakeIndex);
    }

    /// Accept work request with fee
    function requestWork(string calldata requestContentId) external payable {
        require(msg.value >= price, "Insufficient fee");
        uint requestIndex = requestCount++;
        WorkRequest storage request = requests[requestIndex];
        request.customer = msg.sender;
        request.fee = msg.value;
        request.stakeIndex = assignWork(requestIndex);
        request.requestContentId = requestContentId;
    }

    /// Accept work evidence from worker
    function submitWorkEvidence(
        uint requestIndex,
        string calldata evidenceContentId,
        Citation[] calldata citations
    ) external {
        WorkRequest storage request = requests[requestIndex];
        require(
            request.status == WorkStatus.Requested,
            "Status must be Requested"
        );
        AvailabilityStake storage stake = stakes[request.stakeIndex];
        require(
            stake.worker == msg.sender,
            "Worker can only submit evidence for work they are assigned"
        );
        request.status = WorkStatus.EvidenceSubmitted;
        request.evidenceContentId = evidenceContentId;
        for (uint i = 0; i < citations.length; i++) {
            request.citations.push(citations[i]);
        }
        emit WorkEvidenceSubmitted(requestIndex);
    }

    /// Accept work approval/disapproval from customer
    function submitWorkApproval(
        uint requestIndex,
        bool approval
    ) external virtual {
        WorkRequest storage request = requests[requestIndex];
        require(
            request.status == WorkStatus.EvidenceSubmitted,
            "Status must be EvidenceSubmitted"
        );
        AvailabilityStake storage stake = stakes[request.stakeIndex];
        request.status = WorkStatus.ApprovalSubmitted;
        request.approval = approval;
        // Make work evidence post
        Author[] memory authors = new Author[](1);
        authors[0] = Author(100, stake.worker);
        dao.addPost(authors, request.evidenceContentId, request.citations);
        emit WorkApprovalSubmitted(requestIndex, approval);
        // Initiate validation pool
        uint poolIndex = dao.initiateValidationPool{value: request.fee}(
            request.evidenceContentId,
            POOL_DURATION,
            [uint256(1), uint256(3)],
            [uint256(1), uint256(2)],
            100,
            true,
            false,
            ""
        );
        // We have an approval from stake.worker to transfer up to stake.amount
        dao.delegatedStakeOnValidationPool(
            poolIndex,
            stake.worker,
            stake.amount,
            true
        );
    }

    function proposeNewPrice(
        uint newPrice,
        string calldata contentId,
        uint[3] calldata durations
    ) external payable {
        uint priceProposalIndex = priceProposalCount++;
        PriceProposal storage priceProposal = priceProposals[
            priceProposalIndex
        ];
        priceProposal.price = newPrice;
        priceProposal.proposalIndex = proposalsContract.propose{
            value: msg.value
        }(
            contentId,
            msg.sender,
            durations,
            true,
            abi.encode(priceProposalIndex)
        );
        emit PriceChangeProposed(priceProposalIndex);
    }

    function onProposalAccepted(
        uint, // stakedFor
        uint, // stakedAgainst
        bytes calldata callbackData
    ) external returns (uint) {
        uint priceProposalIndex = abi.decode(callbackData, (uint));
        PriceProposal storage priceProposal = priceProposals[
            priceProposalIndex
        ];
        price = priceProposal.price;
        emit PriceChangeAccepted(priceProposalIndex, price);
        return 0;
    }
}
